(****************************************************************************)
(* Level 0/1 Prover                                                         *)
(* Copyright (C) 2004 Alwen Tiu                                             *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

(*
			     CLAUSES.SML
	  The abstract data type for logic program clauses.
*)


signature CLAUSE =
sig
   structure Term : TERM
   
   datatype clause = Clause of Term.term * Term.term 

   val fresh_clause : int * int * clause -> (Term.term * Term.term) 
   val makeClause : Term.term * Term.term * Term.term list -> clause
   val select_clauses : Term.term * clause list -> clause list
end

functor ClauseFunc(
        structure Tm : TERM
        structure Sb : SUBST
        structure Norm : NORM
        sharing Sb.Term = Tm and Norm.Term = Tm) : CLAUSE =
struct    
    structure Term : TERM = Tm
    structure Sub : SUBST = Sb    
    structure Norm : NORM = Norm

	(*
	A clause is a pair of closed terms.
	For example, the clause  p X := q X
	is represented as the pair
	( X\ p X,  X\ q X)
	Notice that free variables in the original clause is
	abstracted in its representation.
	*)
	
	datatype clause = Clause of Term.term * Term.term 
	   	
        fun collapseLam (ref (Term.lam(n, ref (Term.lam(m, t))))) =
		collapseLam (ref (Term.lam(n+m, t)))
          | collapseLam t = t

	fun makeClause(head, body, vars) =
            let val hd = Norm.norm head
                val bd = Norm.norm body 
                val n = length vars
            in
            	if (n = 0) then Clause(hd,bd)
            	else
		  let val h = foldr (fn (v,t) => Term.makeAbsVar(v,t)) hd vars
                      val b = foldr (fn (v,t) => Term.makeAbsVar(v,t)) bd vars
            	  in
                	Clause(collapseLam h, collapseLam b)
            	  end
	    end
	
	
	
	(* create a list of deBruijn indices *)  			
	fun new_dBs 0 = nil
	  | new_dBs n = 
     		(ref (Term.dB n)) :: (new_dBs (n-1))

	(* create new cells for logic variables *)	
	fun new_logicvars i 0 = nil
	  | new_logicvars i n = (Term.freshvar(i))::(new_logicvars i (n-1))
			  			
	fun raise_vars i n m  =
	let val vars = new_logicvars i m
		val dBs = new_dBs n
        in
		map (fn var => if (dBs = nil) then var 
				  else ref (Term.app(var, dBs))) vars 
        end
        
	fun fresh_clause(n, level, Clause(head,body)) =            
            case head of (ref (Term.lam(m, t))) =>
		let val vars = raise_vars level n m
                in
                  (ref (Term.app(head, vars)), ref (Term.app(body, vars)))
                end
              | _ => (head, body)       
              
	fun match(h, ref (Term.const(h',_,_))) = (h = h')
	  | match(h, ref (Term.app(h',_))) = match(h,h')
	  | match(h, ref (Term.lam(n, t))) = match(h, t)
	  | match(h, _) = false
	          
	fun select h nil = nil
		| select h ((c as Clause(head, body))::L) =
			if (match(h, Norm.norm head)) then 
				(c::(select h L))
			else (select h L)
			
	fun select_clauses (A, P) =
		case A of (ref (Term.const(h,_,_))) =>
			select h P
		| ref (Term.app(ref (Term.const(h,_,_)) , _)) =>
			select h P
		| _ => nil
end

                      
