(****************************************************************************)
(* Level 0/1 Prover                                                         *)
(* Copyright (C) 2004 Alwen Tiu                                             *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

(*

lamtree.sml: This module contains the user-level representation of 
the internal lambda terms. 

*)

signature LAMTREE = 
sig
	structure Term : TERM
	exception BadAbstractSyntax

	val abstractSyntax : Term.term -> Term.term
	val abstractDef : Term.term * Term.term -> Term.term
	val concreteSyntax : Term.term -> Term.term
end

functor FuncLamTree (
	structure Term : TERM
	structure Norm : NORM
	sharing Norm.Term = Term) : LAMTREE =
struct

	structure Term = Term
	structure Norm = Norm

	exception BadAbstractSyntax
	
	open Term
	open Norm

	val op_var = ref (Term.const("var",0,none))
	val op_app = ref (Term.const("app",0,none))
	val op_abs = ref (Term.const("abs",0,none))
	val op_const = ref (Term.const("const",0,none))
	
	fun makeConst(s) = ref (const(s,0,none))

	val symTrue = makeConst("symTrue")
	val symFalse = makeConst("symFalse")
	val symAnd = makeConst("symAnd")
	val symOr = makeConst("symOr")
	val symImp = makeConst("symImp")
	val symPi = makeConst("symPi")
	val symSigma = makeConst("symSigma")
	val symNabla = makeConst("symNabla")
	val symIf = makeConst("symIf")
	val symEq = makeConst("symEq")
	val symPrint = makeConst("symPrint")
	val symInput = makeConst("symInput")
	val symParse = makeConst("symParse")
	val symUnparse = makeConst("symUnparse")
	
	fun absSyn (t as ref (var(_,_,_))) = ref (app(op_var, [t]))
	  | absSyn (t as ref (const("def",_,_))) = t
	  | absSyn (t as ref (const("true",_,_))) = symTrue
	  | absSyn (t as ref (const("false",_,_))) = symFalse
	  | absSyn (t as ref (const("&",_,_))) = symAnd
	  | absSyn (t as ref (const(";",_,_))) = symOr
	  | absSyn (t as ref (const("=>",_,_))) = symImp
	  | absSyn (t as ref (const("pi",_,_))) = symPi
	  | absSyn (t as ref (const("sigma",_,_))) = symSigma
	  | absSyn (t as ref (const("nabla",_,_))) = symNabla
	  | absSyn (t as ref (const("if",_,_))) = symIf
	  | absSyn (t as ref (const("=",_,_))) = symEq
	  | absSyn (t as ref (const("print",_,_))) = symPrint
	  | absSyn (t as ref (const("input",_,_))) = symInput
	  | absSyn (t as ref (const("parse",_,_))) = symParse	  
	  | absSyn (t as ref (const("unparse",_,_))) = symUnparse
	  | absSyn (t as ref (const(_,_,_))) = ref (app(op_const, [t]))
	  | absSyn (t as ref (dB(_))) = t
	  | absSyn (ref (app(x, ts))) =
	  		(case ts of nil => absSyn x
	  		     | (y::nil) => ref (app(op_app, [absSyn x, absSyn y]))
	  		     | (s::l) => 
	  		     		let val n = List.length (s::l)
	  		     		    val l1 = List.take (s::l, n-1)
	  		     		    val r = List.last (s::l)
	  		     		    val t = ref (app(x, l1))
	  		     		in
	  		     			ref (app(op_app, [absSyn t, absSyn r]))
	  		     		end 
	  		 )
	  | absSyn (ref (lam(n, t))) =
	  		if (n = 0) then absSyn t
	  		else
	  			let val s = absSyn(ref(lam(n-1,t)))
	  			in
	  				ref (app(op_abs, [ref (lam(1,s))]))
	  			end
	  | absSyn t = raise BadAbstractSyntax
	  
	  	
	fun abstractSyntax t = absSyn(norm t)

	fun makeAbsClause (h,b) = ref (app(ref (const("def",0,none)), [h,b]))

	fun absDef(ref (lam(n,h)), ref (lam(m,b))) = 
		if (m = n) then ref (lam(n, makeAbsClause(h,b)))
		else raise BadAbstractSyntax
          | absDef(h,b) = makeAbsClause(h,b)
		  			
	fun abstractDef(h, b) = absSyn(absDef(h,b))
	
	fun concSyn (t as ref (var(_,_,_))) = t
	  | concSyn (t as ref (const("def",_,_))) = t
	  | concSyn (t as ref (const("symTrue",_,_))) = makeConst("true")
	  | concSyn (t as ref (const("symFalse",_,_))) = makeConst("false")
	  | concSyn (t as ref (const("symAnd",_,_))) = makeConst("&")
	  | concSyn (t as ref (const("symOr",_,_))) = makeConst(";")
	  | concSyn (t as ref (const("symImp",_,_))) = makeConst("=>")
	  | concSyn (t as ref (const("symPi",_,_))) = makeConst("pi")
	  | concSyn (t as ref (const("symSigma",_,_))) = makeConst("sigma")
	  | concSyn (t as ref (const("symNabla",_,_))) = makeConst("nabla")
	  | concSyn (t as ref (const("symIf",_,_))) = makeConst("if")
	  | concSyn (t as ref (const("symEq",_,_))) = makeConst("=")
	  | concSyn (t as ref (const("symPrint",_,_))) = makeConst("print")
	  | concSyn (t as ref (const("symInput",_,_))) = makeConst("input")
	  | concSyn (t as ref (const("symParse",_,_))) = makeConst("parse")	  
	  | concSyn (t as ref (const("symUnparse",_,_))) = makeConst("unparse")	  
	  | concSyn (t as ref (const(_,_,_))) = t
	  | concSyn (t as ref (dB(_))) = t
	  | concSyn (ref (app(h as ref (const(S,_,_)), ts))) = 
	  	(
	  		if (S = "var" orelse S = "const" orelse S = "abs")  then
	  			(
	  				case ts of [v] => concSyn v
	  				| _ => raise BadAbstractSyntax
	  			)
	  		else if (S = "app") then
	  			(
	  				case (map concSyn ts) of nil => raise BadAbstractSyntax
	  				| (t::nil) => t
	  				| (t::rs) => ref (app(t, rs))
	  			)
	  		else 
	  			(
	  			let val rs = map concSyn ts 
	  			in
	  				case rs of nil => h
	  				| _ => ref (app(h, rs))
	  			end
	  			)
		)
	  | concSyn (ref (app(h, ts))) = ref (app(concSyn h, map concSyn ts))
	  | concSyn (ref (lam(n, t))) = ref (lam(n, concSyn t))
	  | concSyn t = raise BadAbstractSyntax
	  
	fun concreteSyntax t = concSyn (norm t)	  
	
end	(* of struct *)
