(****************************************************************************)
(* Level 0/1 Prover                                                         *)
(* Copyright (C) 2004 Alwen Tiu                                             *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

(* Level 0 Prover: Horn-nabla with equality *)

signature PROVER0 =
sig
	structure Core : CORE
	
	val prove : int * Core.Cls.clause list * Core.Term.term list * Core.Sub.subst -> Core.Sub.subst Core.UStream.ustream
end

functor Prover0Func (structure Core : CORE) : PROVER0 = 

struct

structure Core = Core

open Core

structure T = Term


fun body_and_sub(lev, [],_,_,_) = UStream.snil
  | body_and_sub(lev, clause::rest,n,atom,subst) =
    let val (head,body) = 
		Cls.fresh_clause(n, lev, clause) 
        val newhead = Norm.norm(ref (Term.lam(n, ref (Term.app(judg, [head])))))
		val newbody = Norm.norm(ref (Term.lam(n, ref (Term.app(judg, [body])))))
		val (A, newsub) =  Sub.copy_lv atom
    in
      (
        if (is_debug()) then 
		(
            print "Unifying goal : \n";
            PPrint.print_abs_term n A;
            print "\n with the (raised) clause :\n";
		  	PPrint.print_abs_term n newhead; 
		  	print " := ";
		  	PPrint.print_abs_term n newbody;
		  	print "\n"
	    )
		else ();
        (unify A newhead) ;  
      	if (is_debug()) then print "Unification succeeded.\n" else (); 
       	UStream.scons(([newbody], Sub.compose_sub subst newsub),
              (fn ()=>(body_and_sub(lev, rest,n,atom,subst))))
      )
         handle Unif.UniEx(E) =>
         	(
         	    if (is_debug()) then
         	       (print "Unification failed: "; print E; print "\n")
         	    else ();
         		body_and_sub(lev, rest,n,atom,subst)
         	)
         	| Unif.ConstClash(_,_) =>
         	(
         	    if (is_debug()) then
         	       (print "Unification failed: constant clash "; print "\n")
         	    else ();
         		body_and_sub(lev, rest,n,atom,subst)         	
         	)
    end


fun prove_atom (lev, prog, n, A, subst) =
	let val prog' = Cls.select_clauses(A, prog)
		val G = 
			(
				if (is_debug()) then 
					(PPrint.print_term A; print "\n")
				else ();
            	if (n = 0) then ref (Term.app(judg,[A]))
            	else ref (Term.lam(n, ref (Term.app(judg, [A]))))
            )
	in
        UStream.squash (UStream.maps (fn (body,s) => (prove(lev, prog,body,s)))
                    (body_and_sub(lev, prog',n,G,subst))) 
	end
	
and prove_eq (lev, prog, n, E, subst) =
    let val (E', newsub) = Sub.copy_lv E                
    	val E2 = Norm.hnorm E'
    	val (s,t) =
    	    case E2 of ref (Term.app(_, [u, v])) =>
    	            ( ref (Term.lam(n, u)), ref (Term.lam(n,v)))
    	       | _ => raise InvalidGoal
    in
		(unify s t; 
		 UStream.scons(Sub.compose_sub subst newsub,fn () => UStream.snil) )    
    	handle Unif.UniEx(E) => UStream.snil
    		| Unif.ConstClash(_,_) => UStream.snil
    end
    
and prove_disjunct (lev, prog, nil, subst) = UStream.snil
  | prove_disjunct (lev, prog, (G::rest), subst) =
      	UStream.scons(prove(lev, prog, [G], subst), 
      		fn () => prove_disjunct(lev, prog, rest, subst))
      
      
and prove_judgments (lev, prog, n, G, subst) =
	case G of 
	   ref (T.app(ref (Term.const("&", 0, Term.none)), Gs)) => 
	  		(
	  		if (is_debug()) then
	  		 (
	  		 	print "prove conjunction : "; 
	  		 	PPrint.print_list Gs;
	  		 	print "\n"
	  		 )
	  		 else ();
	  		 prove (lev, prog, (make_judgments n Gs), subst)
	  		)
	  |  ref (T.app(ref (Term.const(";", 0, Term.none)), Gs)) => 
	  		(
	  		if (is_debug()) then
	  		 (
	  		 	print "prove disjunction : "; 
	  		 	PPrint.print_list Gs;
	  		 	print "\n"
	  		 )
	  		 else ();
	  		 UStream.squash(prove_disjunct(lev, prog, (make_judgments n Gs), subst))
	  		)	  	

      |  ref (T.app(ref (Term.const("if", 0, Term.none)), [P,Q,R])) => 	  
      		(
      		 print "Level 0 prover: I don't know how to handle if-then-else goal!\n"; 
			 raise InvalidGoal
      		)      			  		
	  | ref (T.app(ref (Term.const("pi", 0, Term.none)), [G' as ref (T.lam(m,_))])) => 
			(print "Level 0 prover: I don't know how to handle universal goal!\n"; 
			 raise InvalidGoal)
	  | ref (T.app(ref (Term.const("nabla", 0, Term.none)), [ref (T.lam(m,G'))])) =>
	  		if (m = 0) then (
	  			if (is_debug()) then print "Empty abstraction!\n" else ();
	  			raise InvalidGoal
	  			)
	  		else prove (lev, prog, (make_judgments (n + m) [G']), subst)

	  | ref (T.app(ref (Term.const("sigma", 0, Term.none)), [G' as ref (T.lam(m,_))])) =>
	  		if (m = 0) then (raise InvalidGoal)
	  		else prove (lev, prog, [(raise_logicvars lev n m G')], subst)
	  | ref (T.app(ref (Term.const("=>", 0, Term.none)), [A, B])) =>
			(print "Level 0 prover: I don't know how to handle implicational goal!\n";
			 raise InvalidGoal)
	  | E as ref (T.app(ref (Term.const("=", 0, Term.none)), [s,t])) =>
	        prove_eq (lev, prog, n, E, subst)
	  |	ref (T.app(ref (Term.const("print", 0, Term.none)), [s])) =>
		    ( 
		       (case s of ref (Term.const(str, _, _)) =>
		       		if (String.isPrefix "\"" str) then
		       		   if (String.size(str) > 2) then
		       		   		print (String.substring(str,1,String.size(str) - 2))
		       		   else ()
		       		else print str
                    | _ =>  PPrint.print_abs_term n s
		    	) ;
		        UStream.scons(subst,fn () => UStream.snil)
		    )	        
	  | ref (Term.const("true", 0, Term.none)) => UStream.scons(subst,fn () => UStream.snil)
	  | ref (Term.const("false", 0, Term.none)) => UStream.snil		        
	  | A as (ref (T.app(ref (T.const(_,_,_)), ts))) =>
	  		(
	  			if (is_debug()) then print "prove atom\n"
	  			else (); 
	  			prove_atom (lev, prog, n, A, subst)  
	  		)
	  | A as (ref (T.const(_,_,_))) =>
	  		(
	  			if (is_debug()) then print "prove atom\n"
	  			else (); 
	  			prove_atom (lev, prog, n, A, subst)  
	  		)	  			  		
	  | _ => (raise InvalidGoal)
  		        
and prove1 (lev, prog, A, subst) =
   let val (ref G) = Norm.norm A
   in
      case G of 
       	  T.app(judg,[G']) => prove_judgments (lev, prog, 0, G', subst)
      	| T.lam(n, ref (T.app(judg,[G']))) =>
      		prove_judgments (lev, prog, n, G', subst)
      	| _ => (PPrint.print_term (ref G) ; raise InvalidGoal)
   end        

and prove(lev, prog, [], subst) = UStream.scons(subst,fn () => UStream.snil)
  | prove(lev, prog, G::rest, subst) =
     UStream.squash
      (UStream.maps
        (fn subst1 =>
            prove(lev, prog,
                  map (fn a => Sub.apply_sub subst1 a) rest,
                  subst1))
       (prove1(lev, prog,G,subst)));


end (* of structure declaration *)
