(****************************************************************************)
(* Level 0/1 Prover                                                         *)
(* Copyright (C) 2004 Alwen Tiu                                             *)
(*                                                                          *)
(* This program is free software; you can redistribute it and/or modify     *)
(* it under the terms of the GNU General Public License as published by     *)
(* the Free Software Foundation; either version 2 of the License, or        *)
(* (at your option) any later version.                                      *)
(*                                                                          *)
(* This program is distributed in the hope that it will be useful,          *)
(* but WITHOUT ANY WARRANTY; without even the implied warranty of           *)
(* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            *)
(* GNU General Public License for more details.                             *)
(*                                                                          *)
(* You should have received a copy of the GNU General Public License        *)
(* along with this code; if not, write to the Free Software Foundation,     *)
(* Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA             *)
(****************************************************************************)

(* subst.ml: Data structure and functions for substitution.                  *)

signature SUBST =
sig
	structure Term : TERM
	
	exception NotPureTerm
	
	datatype subst = Subst of (Term.term * Term.term) list
	
	(* empty substitution *)
	val empty : subst
	
	(* copying a term:
	   
	   copy_lv takes a term t and produce a pair (t', S)
	   where t' is a copy of t, but t' can share subterms
	   with t. S is a substitution (a list of pairs of terms) relating 
	   the variables in t with the corresponding new instances in t'.
	   
	   copy_ev does the same thing for eigenvariables (i.e.,
	   'const' with tag set to ev), but it copies 
	   eigenvariables as variables. 
	   copy_ev can only be applied to closed terms,
	   i.e., terms with no free variables. If this restriction
	   is violated, the exception NotPureTerm is thrown.	   
	*)
	
	val copy_lv : Term.term -> (Term.term * subst)
	val copy_ev : Term.term -> (Term.term * subst)
	
	(* var_to_ev changes all logic variables in a term to eigenvariables.
	   that is, all 'var' nodes are replaced by 'const' nodes
	   tagged with 'ev'. 
	   This function returns the unit value; the update of the term structure
	   is achieved through side-effect.
	*)
	val var_to_ev : Term.term -> unit
	
	val sub_var_to_ev : subst -> unit
	
	(* applying substitution
	   for the moment, assume that susbtitutions are well-formed,
	   i.e., 
	   - all deBruijn numbers are bound by lambda abstractions,
	   - level-restrictions on variables/constants are respected, 
	     that is, a variable can only be substituted by a term
	     of the same or lower level
	   - substitution terms are well-typed and correspond to the
	     type of the variables they substitute for
	     
	 *)
	 
	val apply_sub : subst -> Term.term  -> Term.term
	
	(* applying substitution to a list of terms *)
	val sub_list : subst -> Term.term list -> Term.term list		
		
	(* composition of substitutions *)
	val compose_sub : subst -> subst -> subst
	
end

functor SubstFunc(
	structure Term : TERM
	structure Norm : NORM
	sharing Norm.Term = Term) : SUBST =
struct

	structure Term = Term
	structure Norm = Norm
	
	exception NotPureTerm
	
	datatype subst = Subst of (Term.term * Term.term) list
	
	val empty = Subst nil

	fun get_range x nil = NONE
	  | get_range x ((y,t)::L) =
		if (x = y) then SOME t
		else get_range x L
	
	fun in_domain(x, nil) = false
	  | in_domain(x, ((y,t)::L)) = 
	  	   if (x = y) then true else in_domain(x, L)

	fun cplv (t as (ref (Term.const _))) L = (t, L)	
	  | cplv (t as (ref (Term.var(s,n,f)))) L =
	  		let val newvar = ref (Term.var(s,n,f))
			    val ot = get_range t L
			in
				case ot of
					NONE => (newvar, (t, newvar)::L)
					| (SOME t') => (t', L) 
			end
	   | cplv (t as (ref (Term.dB _))) L = (t, L)
	   | cplv (t as (ref (Term.lam(n, t')))) L =
	    	let val (s, L') = cplv t' L
	    	in
	    		if (s = t') then (t, L)
	    		else (ref (Term.lam(n,s)), L')
	    	end
	  | cplv (t as (ref (Term.app(h, args)))) L =
	  		let val (newh, L1) = cplv h L
	  		    val (newargs, L2) = cplv_list args L1
	  		in
	  			if (newh = h andalso args = newargs) then (t, L)
	  			else (ref (Term.app(newh, newargs)), L2)
	  		end
	  | cplv t L = cplv (Norm.hnorm t) L
	and cplv_list nil L = (nil, L)
	  | cplv_list (t::rs) L = 
	       let val (t', L1) = cplv t L	  		
	           val (rs', L2) = cplv_list rs L1
	       in (t'::rs', L2)  end
	       	      	  	  	
		
	fun copy_lv t =
	    let val (t', L) = cplv t nil
	    in (t', Subst L) end
	 
	
	fun cpev (ref (Term.var _)) L = raise NotPureTerm
	  | cpev (t as (ref (Term.const(s,n,f)))) L =
			if (f = Term.ev) then
  	  			let val newev = ref (Term.var(s,n,f))
			    	val r = get_range t L
			    in
					case r of
						NONE => (newev, (t, newev)::L)
						| (SOME t') => (t', L) 
				end
			else (t, L)
	   | cpev (t as (ref (Term.dB _))) L = (t, L)
	   | cpev (t as (ref (Term.lam(n, t')))) L =
	    	let val (s, L') = cpev t' L
	    	in
	    		if (s = t') then (t, L)
	    		else (ref (Term.lam(n,s)), L')
	    	end
	  | cpev (t as (ref (Term.app(h, args)))) L =
	  		let val (newh, L1) = cpev h L
	  		    val (newargs, L2) = cpev_list args L1
	  		in
	  			if (newh = h andalso args = newargs) then (t, L)
	  			else (ref (Term.app(newh, newargs)), L2)
	  		end
	  | cpev t L = cpev (Norm.hnorm t) L
	and cpev_list nil L = (nil, L)
	  | cpev_list (t::rs) L = 
	       let val (t', L1) = cpev t L	  		
	           val (rs', L2) = cpev_list rs L1
	       in (t'::rs', L2)  end
	       	      
	fun copy_ev t = 
		let val (t', L) = cpev t nil
		in (t', Subst L) end

	    						
	fun var_to_ev (t as (ref (Term.var(s,n,f)))) = (t := Term.const(s,n,Term.ev))
	  | var_to_ev (ref (Term.const _)) = ()
	  | var_to_ev (ref (Term.dB _)) = ()
	  | var_to_ev (ref (Term.lam(_,t))) = var_to_ev t
	  | var_to_ev (ref (Term.app(h,ts))) = 
	  		(var_to_ev h; var_to_ev2 ts)
	  | var_to_ev t = var_to_ev (Norm.hnorm t)
	and var_to_ev2 nil = ()
	  | var_to_ev2 (x::ts) = (var_to_ev x; var_to_ev2 ts)
	  
	fun sub_var_to_ev (Subst nil) = ()
	  | sub_var_to_ev (Subst ((x,t)::rs)) =
	  		(var_to_ev t ; sub_var_to_ev (Subst rs))
	  			
		  
	fun sub_var (Subst nil) x = x
	  | sub_var (Subst ((x,t)::L)) y = 
	  		if (x = y) then t
	  		else sub_var (Subst L) y
	  			  	
	fun apply_sub (Subst nil) t = t
	  | apply_sub S (t as (ref (Term.var _))) = sub_var S t
	  | apply_sub S (t as (ref (Term.const _))) = sub_var S t
	  | apply_sub S (t as (ref (Term.dB n))) = t
	  | apply_sub S (t as (ref (Term.lam(n, t')))) = 
	  		let val s = apply_sub S t'
	  		in
	  			if (s = t') then t
	  			else (ref (Term.lam(n, s)))
	  		end
	  | apply_sub S (t as (ref (Term.app(h, ts)))) =
	  		let val h' = apply_sub S h
	  			val ts' = sub_list S ts
	  		in
	  			if (h = h' andalso ts = ts') then t
	  			else (ref (Term.app(h', ts')))
	  		end	  		
	  | apply_sub S t = apply_sub S (Norm.hnorm t)
	and sub_list (Subst nil) L = nil
	  | sub_list S nil = nil
	  | sub_list S (x::L) = (apply_sub S x)::(sub_list S L)
	  	 	  		

    	    
	local
		fun app_sub (Subst nil) S = S
		  | app_sub S  (Subst nil) = S
		  | app_sub (Subst L) (Subst ((x,t)::L')) =
			let val (Subst L1) = app_sub (Subst L) (Subst L')
			in
				if in_domain(x, L) then (Subst L1)
				else (Subst ((x,t)::L1))
			end
		fun comp (Subst nil) L1 L2 = app_sub L1 L2
		  | comp (Subst ((x,t)::L)) (Subst L1) S = 
			let val t' = apply_sub S t
			    val S' = (Subst ((x,t')::L1))
			in
			  comp (Subst L) S' S
			end
	in
		fun compose_sub (Subst nil) L = L
		  | compose_sub L (Subst nil) = L
		  | compose_sub L1 L2 = comp L1 (Subst nil) L2
	end
			 
end
	
